import asyncio
import base64
import json
from typing import Any, Dict, Optional, Callable, Awaitable, List
import sounddevice as sd
import numpy as np
import threading
import time

# Optional OpenAI realtime imports
try:
    from openai import AsyncOpenAI
    from openai.types.beta.realtime.session import Session
    from openai.resources.beta.realtime.realtime import AsyncRealtimeConnection
    OPENAI_REALTIME_AVAILABLE = True
except ImportError as e:
    OPENAI_REALTIME_AVAILABLE = False
    print(f"OpenAI realtime not available: {e}")
    # Create dummy classes for type hints
    AsyncOpenAI = None
    Session = None
    AsyncRealtimeConnection = None

from dataclasses import dataclass
from datetime import datetime

# --------------------------
# Audio Configuration
# --------------------------
CHUNK_LENGTH_S = 0.05  # 50ms
SAMPLE_RATE = 24000
FORMAT = np.int16
CHANNELS = 1

@dataclass
class TranscriptEntry:
    role: str  # 'user' or 'assistant'
    message: str
    timestamp: datetime = datetime.now()

class AudioPlayerAsync:
    def __init__(self, on_playback_complete: Optional[Callable[[], None]] = None, verbose: bool = False):
        self.queue = []
        self.lock = threading.Lock()
        self.stream = sd.OutputStream(
            callback=self.callback,
            samplerate=SAMPLE_RATE,
            channels=CHANNELS,
            dtype=FORMAT,
            blocksize=int(CHUNK_LENGTH_S * SAMPLE_RATE),
        )
        self.playing = False
        self._frame_count = 0

        # AI speech tracking
        self.is_currently_speaking = False
        self._last_nonzero_timestamp = 0.0
        self.timeout_ai_talking = 0.6

        # Playback-complete callback
        self.on_playback_complete = on_playback_complete
        self._playback_complete_triggered = False

        self.verbose = verbose

    def callback(self, outdata, frames, time_info, status):
        with self.lock:
            data = np.empty(0, dtype=FORMAT)

            # Pull frames from the queue
            while len(data) < frames and len(self.queue) > 0:
                item = self.queue.pop(0)
                needed = frames - len(data)
                data = np.concatenate((data, item[:needed]))
                if len(item) > needed:
                    self.queue.insert(0, item[needed:])

            self._frame_count += len(data)

            # If not enough data, fill with zeros
            if len(data) < frames:
                data = np.concatenate((data, np.zeros(frames - len(data), dtype=FORMAT)))

        # Determine if there's non-silent audio
        if np.any(data != 0):
            self._last_nonzero_timestamp = time.time()

        if time.time() - self._last_nonzero_timestamp < self.timeout_ai_talking:
            self.is_currently_speaking = True
        else:
            self.is_currently_speaking = False

        outdata[:] = data.reshape(-1, 1)

        # Check if the queue is now empty => possible playback completion
        with self.lock:
            # We no longer require 'not self.playing'. We only check empty queue.
            if not self.queue and not self._playback_complete_triggered and self.on_playback_complete:
                # Trigger only once
                self._playback_complete_triggered = True
                self.on_playback_complete()

    def add_data(self, data: bytes):
        """
        Add a chunk of AI audio to the queue.
        Reset the _playback_complete_triggered so that a new callback can fire.
        """
        with self.lock:
            audio_array = np.frombuffer(data, dtype=FORMAT)
            self.queue.append(audio_array)
            if not self.playing:
                self.start()
            # Because we're adding new audio, reset the triggered flag
            self._playback_complete_triggered = False

    def start(self):
        if not self.playing:
            self.playing = True
            self.stream.start()
            print("AudioPlayerAsync: Stream started.")

    def stop(self):
        if self.playing:
            self.playing = False
            self.stream.stop()
            with self.lock:
                self.queue = []
            print("AudioPlayerAsync: Stream stopped.")

    def terminate(self):
        self.stream.close()

class RealTimeVoice:
    def __init__(
        self,
        instructions: str,
        on_user_transcript: Optional[Callable[[str], Awaitable[None]]] = None,
        on_ai_transcript: Optional[Callable[[str], Awaitable[None]]] = None,
        on_ai_audio_complete: Optional[Callable[[], Awaitable[None]]] = None,
        model="gpt-4o-mini-realtime-preview-2024-12-17",
        temperature=0.6,
        max_response_output_tokens="inf",
        voice="alloy",
        mute_mic_while_ai_speaking=True,
        verbose: bool = False,
    ):
        if not OPENAI_REALTIME_AVAILABLE:
            raise ImportError(
                "OpenAI realtime dependencies are not available. "
                "Please install with: pip install openai[realtime] or check your OpenAI library version."
            )
        self.instructions = instructions
        self.on_user_transcript = on_user_transcript
        self.on_ai_transcript = on_ai_transcript
        self.on_ai_audio_complete = on_ai_audio_complete

        self.model = model
        self.temperature = temperature
        self.max_response_output_tokens = max_response_output_tokens
        self.mute_mic_while_ai_speaking = mute_mic_while_ai_speaking
        self.verbose = verbose
        self._mic_muted = False

        # Keep track of transcripts
        self.transcripts: List[TranscriptEntry] = []

        # Audio player with custom callback
        self.client = AsyncOpenAI()
        self.audio_player = AudioPlayerAsync(
            on_playback_complete=self.onAIAudioComplete,
            verbose=self.verbose
        )

        self.REALTIME_API_CONFIG = {
            "modalities": ["text", "audio"],
            "instructions": self.instructions,
            "voice": voice,
            "input_audio_format": "pcm16",
            "output_audio_format": "pcm16",
            "input_audio_transcription": {"model": "whisper-1"},
            "turn_detection": {
                "type": "server_vad",
                "threshold": 0.5,
                "prefix_padding_ms": 500,
                "silence_duration_ms": 1000,
            },
            "tools": [],
            "tool_choice": "auto",
            "temperature": self.temperature,
            "max_response_output_tokens": self.max_response_output_tokens,
        }

        # Threading & control
        self._stop_event = threading.Event()
        self._pause_event = threading.Event()
        self._pause_event.set()

        self._thread: Optional[threading.Thread] = None
        self.conn: Optional[AsyncRealtimeConnection] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._connection_ready = threading.Event()

        # We'll set this True whenever we see 'response.done' that includes audio
        self._audio_complete_pending = False
        self._audio_complete_lock = threading.Lock()

    async def _send_mic_audio(self, connection: AsyncRealtimeConnection) -> None:
        read_size = int(SAMPLE_RATE * 0.02)
        stream = sd.InputStream(channels=CHANNELS, samplerate=SAMPLE_RATE, dtype="int16")
        stream.start()
        print("Microphone recording started.")

        try:
            while not self._stop_event.is_set():
                if not self._pause_event.is_set():
                    await asyncio.sleep(0.1)
                    continue

                if stream.read_available < read_size:
                    await asyncio.sleep(0.01)
                    continue

                data, _ = stream.read(read_size)

                # Mute if AI is speaking or mic is muted
                if self._mic_muted or (self.mute_mic_while_ai_speaking and self.audio_player.is_currently_speaking):
                    data = np.zeros_like(data)

                await connection.input_audio_buffer.append(
                    audio=base64.b64encode(data).decode("utf-8")
                )
                await asyncio.sleep(0)
        except (KeyboardInterrupt, asyncio.CancelledError):
            print("Microphone stream interrupted.")
        finally:
            stream.stop()
            stream.close()
            if not self._stop_event.is_set():
                print("Committing final audio buffer and requesting response.")
                await connection.input_audio_buffer.commit()
                await connection.response.create()
                print("Microphone stream stopped.")

    async def _main_loop(self):
        print("Connecting to the Realtime API...")
        try:
            async with self.client.beta.realtime.connect(model=self.model) as conn:
                self.conn = conn
                self._connection_ready.set()
                if self.verbose:
                    print("Real-time session established.")

                # Update session
                await conn.session.update(session=self.REALTIME_API_CONFIG)
                if self.verbose:
                    print("Session parameters updated.")

                mic_task = asyncio.create_task(self._send_mic_audio(conn))
                acc_items: Dict[str, str] = {}

                try:
                    async for event in conn:
                        if self._stop_event.is_set():
                            if self.verbose:
                                print("Stop event detected. Exiting event loop.")
                            break
                        if not self._pause_event.is_set():
                            continue

                        if event.type == "session.created":
                            if self.verbose:
                                print(f"Session created with ID: {event.session.id}")

                        elif event.type == "session.updated":
                            if self.verbose:
                                print("Session updated.")

                        elif event.type == "response.audio.delta":
                            chunk_bytes = base64.b64decode(event.delta)
                            self.audio_player.add_data(chunk_bytes)

                        elif event.type == "response.audio_transcript.delta":
                            item_id = event.item_id
                            if item_id not in acc_items:
                                acc_items[item_id] = event.delta
                            else:
                                acc_items[item_id] += event.delta

                        elif event.type == "conversation.item.input_audio_transcription.completed":
                            user_message = event.transcript
                            self.transcripts.append(TranscriptEntry(role="user", message=user_message))
                            if self.on_user_transcript:
                                asyncio.create_task(self.on_user_transcript(user_message))

                        elif event.type == "response.done":
                            # The model finished an output turn
                            # If there's audio in that response, we want the playback-complete callback
                            if event.response and event.response.output:
                                for item in event.response.output:
                                    if item.type == "message":
                                        for content in item.content:
                                            if content.type == "audio":
                                                # We got AI audio
                                                ai_message = content.transcript
                                                self.transcripts.append(
                                                    TranscriptEntry(role="ai", message=ai_message)
                                                )
                                                if self.on_ai_transcript:
                                                    asyncio.create_task(self.on_ai_transcript(ai_message))
                                                # Mark that we expect an audio-complete event
                                                with self._audio_complete_lock:
                                                    self._audio_complete_pending = True
                        else:
                            if self.verbose:
                                print(f"Unhandled event type: {event.type}")  # Print unhandled events
                except asyncio.CancelledError:
                    if self.verbose:
                        print("Main loop cancelled.")
                    print("Main loop has been cancelled.")
        except Exception as e:
            if self.verbose:
                print(f"An error occurred in the main loop: {e}")
        finally:
            print("Connection closed. Exiting.")

    def _run_loop(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        try:
            self._loop.run_until_complete(self._main_loop())
        except Exception as e:
            print(f"Exception in run loop: {e}")
        finally:
            self._loop.close()
            print("Asyncio event loop closed.")

    def start(self):
        if self._thread and self._thread.is_alive():
            if self.verbose:
                print("RealTimeVoice is already running.")
            return
        if self.verbose:
            print("Starting RealTimeVoice...")
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()
        if self.verbose:
            print("RealTimeVoice started.")

    def stop(self):
        print("Stopping RealTimeVoice...")
        self._stop_event.set()

        if self.conn:
            print("Closing the connection...")
            future = asyncio.run_coroutine_threadsafe(self.conn.close(), self._loop)
            try:
                future.result(timeout=5)
                print("Connection closed successfully.")
            except Exception as e:
                print(f"Error closing connection: {e}")

        if self._thread and self._thread.is_alive():
            print("Waiting for the thread to finish...")
            self._thread.join(timeout=10)
            if self._thread.is_alive():
                print("Thread did not finish in time.")
            else:
                print("Thread has been joined successfully.")

        self.audio_player.stop()
        print("RealTimeVoice stopped.")

    def pause(self):
        print("Pausing RealTimeVoice...")
        self._pause_event.clear()

    def resume(self):
        print("Resuming RealTimeVoice...")
        self._pause_event.set()

    def mute_mic(self):
        self._mic_muted = True
        if self.verbose:
            print("Microphone muted.")

    def unmute_mic(self):
        self._mic_muted = False
        if self.verbose:
            print("Microphone unmuted.")

    def inject_message(self, message: str):
        async def _inject():
            if not self.conn:
                print("No connection available for injecting message.")
                return
            try:
                trigger = {
                    "type": "conversation.item.create",
                    "item": {
                        "type": "message",
                        "role": "user",
                        "content": [{"type": "input_text", "text": message}],
                    },
                }
                await self.conn.send(trigger)
                await self.conn.send({"type": "response.create", "response": {"modalities": ["audio", "text"]}})
                print(f"Injected message: {message}")
            except Exception as e:
                print(f"Error injecting message: {e}")

        if not self._connection_ready.is_set():
            print("Waiting for connection to be ready to inject message...")
            self._connection_ready.wait()

        if not self._loop or not self._loop.is_running():
            print("Event loop not running. Cannot inject message.")
            return

        future = asyncio.run_coroutine_threadsafe(_inject(), self._loop)
        try:
            future.result(timeout=5)
        except Exception as e:
            print(f"Error scheduling inject_message coroutine: {e}")

    def update_instructions(self, new_instructions: str):
        async def _update():
            if not self.conn:
                print("No connection available for updating instructions.")
                return
            try:
                self.instructions = new_instructions
                updated_config = dict(self.REALTIME_API_CONFIG)
                updated_config["instructions"] = self.instructions
                await self.conn.session.update(session=updated_config)
                print(f"Session instructions updated to: {new_instructions}")
            except Exception as e:
                print(f"Error updating instructions: {e}")

        if not self._connection_ready.is_set():
            print("Waiting for connection to be ready to update instructions...")
            self._connection_ready.wait()

        if not self._loop or not self._loop.is_running():
            print("Event loop not running. Cannot update instructions.")
            return

        future = asyncio.run_coroutine_threadsafe(_update(), self._loop)
        try:
            future.result(timeout=5)
        except Exception as e:
            print(f"Error scheduling update_instructions coroutine: {e}")

    def onAIAudioComplete(self):
        """
        Called by AudioPlayerAsync once the queue is empty.
        We'll fire our on_ai_audio_complete callback only if we know
        the AI actually produced audio (i.e., _audio_complete_pending).
        """
        with self._audio_complete_lock:
            if self._audio_complete_pending and self.on_ai_audio_complete:
                if self.verbose:
                    print("Playback complete and AI audio complete callback triggered.")
                self._audio_complete_pending = False  # consume the flag
                if self._loop and self._loop.is_running():
                    async def _callback():
                        await self.on_ai_audio_complete()

                    asyncio.run_coroutine_threadsafe(_callback(), self._loop)
                else:
                    print("Event loop not running. Cannot execute on_ai_audio_complete callback.")

# --------------------------
# Example Usage
# --------------------------
if __name__ == "__main__":
    if not OPENAI_REALTIME_AVAILABLE:
        print("OpenAI realtime dependencies are not available. Cannot run example.")
        exit(1)
        
    instructions = "Respond briefly and with a sarcastic attitude."
    temperature = 0.9
    voice = "echo"
    mute_mic_while_ai_speaking = True
    verbose = False  # Enable verbose prints

    # Optional: callback for when the whisper transcription is done
    async def on_user_transcript(transcript: str):
        print(f"(on_user_transcript) User said: {transcript}")

    # Optional: callback for when the transcript of the voice response is there
    async def on_ai_transcript(transcript: str):
        print(f"(on_ai_transcript) AI replied: {transcript}")

    # Optional: callback for when the audio has been completely played
    async def on_audio_complete():
        print("(on_audio_complete) AI audio has been completely played.")

    rtv = RealTimeVoice(
        instructions=instructions,
        on_user_transcript=on_user_transcript,
        on_ai_transcript=on_ai_transcript,
        on_audio_complete=on_audio_complete,
        model="gpt-4o-mini-realtime-preview-2024-12-17",
        temperature=temperature,
        voice=voice,
        mute_mic_while_ai_speaking=mute_mic_while_ai_speaking,
        max_response_output_tokens="inf",
        verbose=verbose,  # Pass verbose flag
    )

    rtv.start()
    rtv.inject_message("Hello AI, what's up?")

    try:
        while True:
            cmd = input("Commands: (p) pause, (r) resume, (s) stop, (i) inject <msg>, (u) update_instructions <text>, (t) print_transcript\n> ").strip()
            if cmd.lower() == "p":
                rtv.pause()
            elif cmd.lower() == "r":
                rtv.resume()
            elif cmd.lower() == "s":
                rtv.stop()
                break
            elif cmd.lower().startswith("i "):
                message = cmd[len("i "):].strip()
                rtv.inject_message(message)
            elif cmd.lower().startswith("u "):
                new_instructions = cmd[len("u "):].strip()
                rtv.update_instructions(new_instructions)
            elif cmd.lower() == "t":
                print("\n".join([f"{entry.timestamp} {entry.role}: {entry.message}" for entry in rtv.transcripts]))
            else:
                print("Unknown command.")
    except KeyboardInterrupt:
        rtv.stop()
        print("\nExiting.")



from setuptools import setup, find_packages
import os

# Read requirements.txt if it exists (for backward compatibility)
# When using pyproject.toml, dependencies are read from there instead
required = []
if os.path.exists('requirements.txt'):
    with open('requirements.txt') as f:
        required = f.read().splitlines()

# Read README for long description
long_description = ''
if os.path.exists('README.md'):
    with open('README.md') as f:
        long_description = f.read()

setup(
    name='lunar-tools',
    version='0.3',
    packages=find_packages(),
    package_data={
        'lunar_tools': ['midi_configs/*.yml'],
    },
    url='https://github.com/lunarring/lunar_tools',
    description='Lunar Ring auxiliary tools and modules for programming interactive exhibitions',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.10',
    install_requires=required,
    extras_require={
        'torch': ['torch>=2.0.0'],
    },
    include_package_data=True,
)



import logging
import sys

# ANSI escape sequences for colors
COLORS = {
    "black": "\u001b[30;1m",
    "red": "\u001b[31;1m",
    "green": "\u001b[32;1m",
    "yellow": "\u001b[33;1m",
    "blue": "\u001b[34;1m",
    "magenta": "\u001b[35;1m",
    "cyan": "\u001b[36;1m",
    "white": "\u001b[37;1m",
    "reset": "\u001b[0m"
}

try:
    import shutil
except ImportError:  # pragma: no cover - shutil is part of stdlib but keep fallback
    shutil = None

DEFAULT_TERMINAL_WIDTH = 80


def _get_terminal_width():
    """Return current terminal width or a sensible default."""
    if shutil is None:
        return DEFAULT_TERMINAL_WIDTH
    get_size = getattr(shutil, "get_terminal_size", None)
    if get_size is None:
        return DEFAULT_TERMINAL_WIDTH
    try:
        return get_size(fallback=(DEFAULT_TERMINAL_WIDTH, 0)).columns
    except OSError:
        return DEFAULT_TERMINAL_WIDTH


def dynamic_print(message, *, stream=None, persist=False):
    """
    Dynamically prints a message to the console, replacing the previous message.
    Handles messages that are longer than the terminal width by clearing the line first
    and truncating the message if necessary. Also handles multi-line messages by
    converting them to a single line.

    Args:
        message: Text to display.
        stream: Optional stream to write to (defaults to sys.stdout).
        persist: When True the message is followed by a newline so it remains on screen.
    """
    stream = stream or sys.stdout
    text = str(message).replace("\n", " ").replace("\r", "")
    is_tty = bool(getattr(stream, "isatty", lambda: False)())

    if not is_tty:
        stream.write(f"{text}\n")
        stream.flush()
        return

    terminal_width = max(_get_terminal_width(), 1)

    if terminal_width > 3 and len(text) > terminal_width - 3:
        text = text[: terminal_width - 3] + "..."
    elif len(text) > terminal_width:
        text = text[:terminal_width]

    # Clear the current line completely
    stream.write("\r" + " " * terminal_width)
    stream.flush()

    # Return to the beginning of the line and write the new message
    stream.write("\r" + text)
    if persist:
        stream.write("\n")
    stream.flush()

class LogPrint:
    """Light-weight console logger with optional file logging and ANSI colours."""

    LEVEL_MAP = {
        1: logging.DEBUG,
        2: logging.INFO,
        3: logging.CRITICAL,
    }

    def __init__(self, filename=None, verbose_level_console=3, file_mode="a"):
        """
        Args:
            filename: Optional path for file logging. When omitted only console output is used.
            verbose_level_console: Minimum verbosity to show in the console (1=DEBUG, 2=INFO, 3=CRITICAL).
            file_mode: Mode for `logging.FileHandler`. Defaults to append to avoid truncation.
        """
        self.verbose_level_console = verbose_level_console
        self.filename = filename
        self._file_mode = file_mode
        self.logger = None

        if filename:
            # Create a dedicated logger instance so multiple LogPrint objects don't share handlers.
            logger_name = f"logprint.{id(self)}"
            self.logger = logging.getLogger(logger_name)
            self.logger.setLevel(logging.DEBUG)
            self.logger.propagate = False
            # Remove handlers that could remain from stale instances (e.g. during tests)
            self.logger.handlers.clear()

            file_handler = logging.FileHandler(self.filename, mode=self._file_mode, encoding="utf-8")
            file_handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
            self.logger.addHandler(file_handler)

    def print(self, message, color=None, verbose_level=3):
        """Print ``message`` to the console (optionally coloured) and log to file if configured."""
        level = self.LEVEL_MAP.get(verbose_level, logging.CRITICAL)

        if verbose_level >= self.verbose_level_console:
            if color and color in COLORS:
                sys.stdout.write(f"{COLORS[color]}{message}{COLORS['reset']}\n")
            else:
                sys.stdout.write(f"{message}\n")
            sys.stdout.flush()

        if self.logger:
            self.logger.log(level, message)

if __name__ == "__main__":
    # Example usage
    logger = LogPrint()  # No filename provided, will use default current_dir/logs/%y%m%d_%H%M
    logger.print("This DEBUG message is a white message that won't be shown in console", verbose_level=1)
    logger.print("This INFO message is a white message that won't be shown in console", verbose_level=2)
    logger.print("This CRITICAL message is a red message", "red")
    logger.print("This CRITICAL is a green message", "blue")

    import time
    for i in range(100):
        dynamic_print(f"This {i} DEBUG message is a white message that won't be shown in consoleThis DEBUG message is a white message that won't be shown in consoleThis DEBUG message is a white message that won't be shown in consoleThis is a message {i}")
        time.sleep(0.1)



try:
    from pygame import midi
except Exception as e:
    print(f"IMPORT FAIL: {e}")
import numpy as np
import time
import os
import yaml
import threading
import pkg_resources
import inspect
import json
import subprocess
import re
import usb.core     # pip install pyusb
import platform
from datetime import datetime


def get_midi_device_vendor_product_ids(system_device_name):
    # Initialize the result dictionary
    vendor_product_ids = {}

    try:
        os_system = platform.system()
        if os_system == 'Linux':
            # Run the lsusb command for Linux
            usb_output = subprocess.check_output(['lsusb'], text=True)
            regex = f'ID (\\w+:\\w+).+{system_device_name}'
        elif os_system == 'Darwin':
            # Run the system_profiler command for macOS
            usb_output = subprocess.check_output(['system_profiler', 'SPUSBDataType'], text=True)
            regex = f'{system_device_name}.*?\\n.*?Product ID: (0x\\w+)\\n.*?Vendor ID: (0x\\w+)'
        elif os_system == 'Windows':
            # Use WMIC command to retrieve USB device details on Windows
            usb_output = subprocess.check_output(['wmic', 'path', 'Win32_USBHub', 'get', 'DeviceID'], text=True, stderr=subprocess.DEVNULL)
            regex = f'{system_device_name}.*?VID_([0-9A-Fa-f]{{4}})&PID_([0-9A-Fa-f]{{4}})'
        else:
            print("Unsupported operating system.")
            return vendor_product_ids

        # Find all matches
        matches = re.findall(regex, usb_output, re.IGNORECASE | re.DOTALL)

        for match in matches:
            if os_system == 'Linux':
                # Split the match into vendor and product IDs for Linux
                vendor_id, product_id = match.split(':')
            elif os_system == 'Darwin':
                # Assign vendor and product IDs for macOS
                product_id, vendor_id = match
            elif os_system == 'Windows':
                # On Windows, match returns VID and PID groups
                vendor_id, product_id = match

            # Convert hexadecimal to integer and add to the dictionary
            vendor_product_ids = {'vendor_id': int(vendor_id, 16), 'product_id': int(product_id, 16)}
            break

        return vendor_product_ids

    except Exception as e:
        print(f"An error occurred: {e}")
        return vendor_product_ids

    
def check_midi_device_connected_pyusb(device_code):
    if len(device_code) > 0:
        # Look for all devices connected
        devices = usb.core.find(find_all=True)
    
        # Iterate through each device and check if <device_code> device connected
        for device in devices:
            if device.idVendor == device_code['vendor_id'] and device.idProduct == device_code['product_id']:
                return True

    return False    


def check_any_midi_device_connected():
    config_path = pkg_resources.resource_filename('lunar_tools', 'midi_configs')
    for filename in os.listdir(config_path):
        if filename.endswith(".yml"):
            with open(os.path.join(config_path, filename), 'r') as file:
                config = yaml.safe_load(file)
                device_name = config['name_device']
                device_ids = get_midi_device_vendor_product_ids(device_name)
                if len(device_ids) > 0:
                    return filename.split(".yml")[0]
    return None

#%%
class MidiInput:
    """ A class to track midi inputs. """
    def __init__(self,
                 device_name="akai_lpd8",
                 allow_fail=True,
                 device_id_input=None,
                 device_id_output=None,
                 enforce_local_config=False,
                 do_auto_reconnect=False
                 ):

        # get operating system
        if platform.system() == 'Linux':
            self.os_name = 'linux'
        elif platform.system() == 'Darwin':
            self.os_name = 'macos'
        elif platform.system() == 'Windows':
            self.os_name = 'windows'
        else:
            raise NotImplementedError("Only Linux, MacOS and Windows supported at the moment.")            
        
        self.do_auto_reconnect = do_auto_reconnect
        self.simulate_device = False
        self.device_name = device_name
        self.allow_fail = allow_fail
        self.device_id_input = device_id_input
        self.device_id_output = device_id_output
        self.init_device_config(enforce_local_config)
        self.init_device_hardware_code()
        self.init_vars()
        self.init_midi()
        self.reset_all_leds()
        self.autodetect_varname = True
        self.autoshow_names = True
        self.valid_button_modes = ['held_down', 'released_once', 'toggle', 'pressed_once']
        
        
    def init_device_config(self, enforce_local_config):
        # Determine the path to the YAML file
        config_filename = f"{self.device_name}.yml"
        if enforce_local_config:
            config_path = os.path.join("midi_configs", config_filename)
        else:
            config_path = pkg_resources.resource_filename('lunar_tools', f'midi_configs/{config_filename}')

        # Load the configuration from the YAML file
        with open(config_path, 'r') as file:
            config = yaml.safe_load(file)

        # Set the device configuration from the YAML file
        self.system_device_name = config['name_device']
        self.button_down = config['button_down']
        self.button_release = config['button_release']
        self.id_config = config['controls']

        # Reverse for last lookup
        self.reverse_control_name = {(v[0], v[1]): k for k, v in self.id_config.items()}
        
    def init_device_hardware_code(self):
        self.device_hardware_code = get_midi_device_vendor_product_ids(self.system_device_name)

    def init_vars(self):
        # Initializes all variables
        self.id_value = {}
        self.id_last_time_scanned = {}
        self.id_last_time_retrieved = {}
        self.id_nmb_button_down = {}
        self.id_nmb_scan_cycles = {}
        self.id_name = {}
        self.id_val_min = {}  # Store val_min for each control
        self.id_val_max = {}  # Store val_max for each control
        self.id_val_default = {}  # Store val_default for each control
        self.released_once_flags = {}  # Adding a new dictionary to track released_once flags
        for key in self.id_config:
            control_type = self.id_config[key][1]
            self.released_once_flags[key] = False 
            self.id_value[key] = False if control_type == "button" else 0.0
            self.id_last_time_scanned[key] = 0
            self.id_last_time_retrieved[key] = 0
            self.id_nmb_button_down[key] = 0 if control_type == "button" else None
            self.id_nmb_scan_cycles[key] = 0
            # Initialize scaling parameters
            if control_type == "slider":
                self.id_val_min[key] = 0
                self.id_val_max[key] = 1
                self.id_val_default[key] = 0.5
            else:
                self.id_val_min[key] = None
                self.id_val_max[key] = None
                self.id_val_default[key] = False

    def compare_device_names(self, dev_info):
        if self.system_device_name in dev_info[1].decode():
            return True
        elif dev_info[1].decode() in self.system_device_name:
            return True
        else:
            return False
        
    def auto_determine_device_id(self, is_input):
        dev_count = midi.get_count()
        device_id = None
        for i in range(dev_count):
            dev_info = midi.get_device_info(i)
            if self.compare_device_names(dev_info):
                if is_input and dev_info[2] == 1:
                    device_id = i
                elif not is_input and dev_info[3] == 1:
                    device_id = i
        if not self.allow_fail:
            assert device_id, f"Could not auto determine device_id for {is_input}"
        if is_input:
            self.device_id_input = device_id
        else:
            self.device_id_output = device_id
        
    def check_device_id(self, is_input):
        if self.allow_fail and self.device_id_input is None:
            self.simulate_device = True
            print("Simulating midi device!")
            return
        elif not self.allow_fail and self.device_id_input is None:
            raise ValueError("Device init failed! If you want to simulate, set allow_fail=True")
        if is_input:
            dev_info = midi.get_device_info(self.device_id_input)
        else:
            dev_info = midi.get_device_info(self.device_id_output)
        
        if not self.compare_device_names(dev_info):
            print(f"Device mismatch: name_device={self.system_device_name} and get_device_info={dev_info[1].decode()}")
            return False
        else:
            return True
        
        
    def init_midi(self):
        # Have a small safety loop for init
        for i in range(5):
            midi.quit()
            time.sleep(0.01)
            midi.init()
        
        # Set the device_ids
        if self.device_id_input is None:
            self.auto_determine_device_id(is_input=True)
        if self.device_id_output is None:
            self.auto_determine_device_id(is_input=False)
            
        # Check the device_ids
        is_valid_input = self.check_device_id(is_input=True)
        is_valid_output = self.check_device_id(is_input=False)
        if not self.simulate_device: 
            assert is_valid_input
            assert is_valid_output
        
        # Init midi in and out
        if not self.simulate_device:
            self.midi_in = midi.Input(self.device_id_input)
            self.midi_out = midi.Output(self.device_id_output)

    def get_control_name(self, idx_control, type_control):
        if type_control == self.button_down or type_control == self.button_release:
            str_type_control = "button"
        else:
            str_type_control = "slider"
        
        if (idx_control, str_type_control) in self.reverse_control_name:
            return self.reverse_control_name[(idx_control, str_type_control)]
        else:
            return None

    def scan_inputs(self):
        if self.simulate_device:
            return
        
        # Gather all inputs that arrived in the meantime
        while True:
            if self.os_name in ['linux', 'windows'] and self.do_auto_reconnect:
                time.sleep(1e-3)
                is_midi_device_connected = check_midi_device_connected_pyusb(self.device_hardware_code)
                print(f'{is_midi_device_connected}')
                if not is_midi_device_connected:
                    print(f'{self.device_name} has disconnected. trying to reconnect...')
                    self.init_midi()                
            else:
                is_midi_device_connected = True
                
            if is_midi_device_connected:
                input_last = self.midi_in.read(1)
            else:
                break
                
            if input_last == []:
                break
            type_control = input_last[0][0][0]
            idx_control = input_last[0][0][1]
            val_control = input_last[0][0][2]
            
            alpha_num = self.get_control_name(idx_control, type_control)
            
            # Process the inputs
            if self.id_config[alpha_num][1] == "slider":
                self.id_value[alpha_num] = val_control / 127.0
                self.id_last_time_scanned[alpha_num] = time.time()
                
                
            elif self.id_config[alpha_num][1] == "button":
                if type_control == self.button_down:
                    self.id_last_time_scanned[alpha_num] = time.time()
                    self.id_nmb_button_down[alpha_num] += 1
                    self.id_value[alpha_num] = True
                else:
                    self.id_value[alpha_num] = False
                    self.released_once_flags[alpha_num] = True 
    
            
    def get(self, alpha_num, val_min=None, val_max=None, val_default=False, button_mode=None, variable_name=None):
        """
        Retrieves the value of a control input based on its alphanumeric identifier.

        Parameters:
        - alpha_num (str): The alphanumeric identifier of the control input.
        - val_min (float, optional): The minimum value for sliders. Defaults to None.
        - val_max (float, optional): The maximum value for sliders. Defaults to None.
        - val_default (bool or float, optional): The default value for the control input. For buttons, it's a boolean indicating pressed state; for sliders, it's their position. Defaults to False.
        - button_mode (str, optional): The mode of button operation. Can be one of ['held_down', 'released_once', 'toggle', 'pressed_once']. Defaults to None.
        - variable_name (str, optional): The name of the variable to override the autodetection. Defaults to None.

        Returns:
        - The current value of the control input based on its configuration and the provided parameters.
        """
        # Asserts
        if alpha_num not in self.id_config:
            print(f"Warning! {alpha_num} is unknown. Returning val_default")
            return val_default
        # Assert that button mode is correct if button
        if val_default and button_mode is None:
            assert val_default >= val_min and val_default <= val_max
        # Assert correct type
        if self.id_config[alpha_num][1] == "button":
            assert val_min is None and val_max is None, f"{alpha_num} is a button cannot have val_min/val_max"
        if self.id_config[alpha_num][1] == "slider":
            assert button_mode is None, f"{alpha_num} is a slider cannot have button_mode"
        
        # Set default args
        if val_min is None:
            val_min = 0
        if val_max is None:
            val_max = 1
        if button_mode is None:
            button_mode = 'released_once'
        assert button_mode in self.valid_button_modes
        
        # Store scaling parameters for later use in get_parameters_dict
        if alpha_num in self.id_config:
            if self.id_config[alpha_num][1] == "slider":
                self.id_val_min[alpha_num] = val_min
                self.id_val_max[alpha_num] = val_max
                self.id_val_default[alpha_num] = val_default
            else:  # button
                self.id_val_default[alpha_num] = val_default
        
        if self.autodetect_varname and variable_name is None:
            if self.id_nmb_scan_cycles[alpha_num] <= 2:
                # Inspecting the stack to find the variable name
                frame = inspect.currentframe()
                try:
                    outer_frame = frame.f_back
                    call_line = outer_frame.f_lineno
                    source_lines, starting_line = inspect.getsourcelines(outer_frame)
                    line_index = call_line - starting_line
                    call_code = source_lines[line_index-1].strip()
        
                    # Extracting the variable name
                    variable_name = call_code.split('=')[0].strip()
                except Exception:
                    variable_name = "autodetect failed"
                finally:
                    del frame  # Prevent reference cycles
                
                if self.id_nmb_scan_cycles[alpha_num] == 1:
                    assert variable_name == self.id_name[alpha_num], f"Double assignment for {alpha_num}: {variable_name} and {self.id_name[alpha_num]}"
                self.id_nmb_scan_cycles[alpha_num] += 1
                self.id_name[alpha_num] = variable_name
            
        elif variable_name is not None:
            self.id_name[alpha_num] = variable_name
    
        if self.id_nmb_scan_cycles[alpha_num] == 2 and self.autoshow_names:
            self.show()
            self.autoshow_names = False
        # Scan new inputs
        try:
            # Auto device reconnect/disconnect handling for Linux and Windows
            self.scan_inputs()
        except Exception as e:
            print(f"scan_inputs raised: {e}")
        
        # Process slider
        if self.id_config[alpha_num][1] == "slider":
            if val_default is False:
                val_default = 0.5 * (val_min + val_max)
            if self.id_last_time_scanned[alpha_num] == 0:
                val_return = val_default
            else:
                val_return = val_min + (val_max-val_min) * self.id_value[alpha_num]
        
        # Process button
        elif self.id_config[alpha_num][1] == "button":
            if val_default:
                assert button_mode == "toggle", "if val_default is set True, button_mode must be 'toggle'"
            if button_mode == 'held_down':
                val_return = self.id_value[alpha_num]
            elif button_mode == "pressed_once":
                val_return = self.id_value[alpha_num] and self.id_last_time_scanned[alpha_num] > self.id_last_time_retrieved[alpha_num]
                if val_return:
                    self.id_value[alpha_num] = False  # Reset the value after being processed
            elif button_mode == "released_once":
                val_return = self.released_once_flags.get(alpha_num, False)
                if val_return:
                    self.released_once_flags[alpha_num] = False # Reset the last time scanned after being processed
            elif button_mode == "toggle":
                if val_default is None:
                    val_default = False
                val_return = np.mod(self.id_nmb_button_down[alpha_num]+1, 2) == val_default
                # Set LED
                self.set_led(alpha_num, val_return)
                
        self.id_last_time_retrieved[alpha_num] = time.time()
        
        return val_return
    
    def get_parameters_dict(self, **misc):
        """ Return a dictionary with all assigned parameters and their values. Additional information can be passed as kwargs"""
        parameters = []
        for id_, name in self.id_name.items():
            if self.id_config[id_][1] == "slider":
                # Scale the raw value using stored min/max parameters
                if self.id_last_time_scanned[id_] == 0:
                    # No input received yet, use default value
                    scaled_value = self.id_val_default[id_]
                else:
                    # Scale the raw value (0-1) to the desired range
                    val_min = self.id_val_min[id_]
                    val_max = self.id_val_max[id_]
                    scaled_value = val_min + (val_max - val_min) * self.id_value[id_]
                parameters.append({'id': id_, 'name': name, 'value': scaled_value})
            else:
                # For buttons, return the boolean value as-is
                value = self.id_value[id_]
                parameters.append({'id': id_, 'name': name, 'value': value})
        if misc:
            parameters.append(misc)
        return parameters    

        
    def set_led(self, alpha_num, state):
        if self.simulate_device:
            return
        assert alpha_num in self.id_config
        assert self.id_config[alpha_num][1] == "button"
        self.midi_out.write([[[self.button_down, self.id_config[alpha_num][0], state, 0], 0]])
        
    def reset_all_leds(self):
        for alpha_num in self.id_config:
            if self.id_config[alpha_num][1] == "button":
                self.set_led(alpha_num, False)
                
    def show(self):
        """
        shows the assignemnet of the alpha_nums on the midi device
        """
        # Extract letters and numbers
        letters = sorted(set(key[0] for key in self.id_config.keys()))
        max_num = max(int(key[1]) for key in self.id_config.keys())
        
        # Determine the maximum width of each column
        max_widths = {letter: max(len(self.id_name.get(f"{letter}{num}", '-')) for num in range(max_num + 1)) for letter in letters}
        
        # Print header row
        header_row = '   ' + ' '.join(letter.center(max_widths[letter]) for letter in letters)
        print('\n')
        print(header_row)
        print('   ' + '+'.join('-' * max_widths[letter] for letter in letters))
        
        # Create the grid with left header
        for num in range(max_num + 1):
            row = f"{num} |"
            for letter in letters:
                key = f"{letter}{num}"
                row += self.id_name.get(key, '-').center(max_widths[letter]) + '|'
            print(row)
        print('\n')
        
    def print_values(self):
        print('id\tname\tvalue\n')
        for id_, name in self.id_name.items():
            value = self.id_value[id_]
            print(f'{id_}\t{name}\t{value}')
        print('\n')

       


#%%

       
                    
if __name__ == "__main__":
    import time
    akai_lpd8 = MidiInput(device_name="akai_lpd8")
    
    while True:
        time.sleep(0.2)
        # released_once checks if the button was pressed since we checked last time
        z = akai_lpd8.get("A0", button_mode='toggle', val_default=True) 
        # released_once checks if the button was pressed since we checked last time
        x = akai_lpd8.get("B0", button_mode='released_once') 
        #  e0 is a slider float between val_min and val_max
        slider = akai_lpd8.get("E0", val_min=3, val_max=6, val_default=5) #
        print(f"z: {z}, x: {x} slider: {slider}")
        
    akai_lpd8.show()
                    
if __name__ == "__main__":
    import lunar_tools as lt
    import time
    akai_midimix = MidiInput(device_name="akai_midimix")
    
    
    while True:
        time.sleep(0.1)
        a0 = akai_midimix.get("A0", val_min=3, val_max=6, val_default=5)
        ba = akai_midimix.get("A3", button_mode="toggle")
        print(ba)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import unittest
from unittest import mock

from lunar_tools.logprint import LogPrint, dynamic_print

class TestLogPrint(unittest.TestCase):

    def test_log_creation_with_specific_filename(self):
        """Test if a log file is created with a specific filename."""
        specific_filename = "tmp_testlog.txt"
        logger = LogPrint(filename=specific_filename)
        logger.print("Test log with specific filename.")
        self.assertTrue(os.path.exists(specific_filename))

    def test_log_creation_with_default_filename(self):
        """When no filename is provided, LogPrint should avoid file logging."""
        logger = LogPrint()
        logger.print("Test log with default filename.")
        self.assertIsNone(logger.filename)

    def test_log_file_content(self):
        """Test if the log file contains the expected content."""
        test_filename = "specific_filename.txt"
        test_message = "This is a test log message."
        logger = LogPrint(filename=test_filename)
        logger.print(test_message)
        with open(test_filename, 'r') as file:
            content = file.read()
            self.assertIn(test_message, content)

    def test_dynamic_print_persist_writes_newline(self):
        """dynamic_print should append a newline when persist=True."""
        class FakeTTYStream(io.StringIO):
            def isatty(self):
                return True

        stream = FakeTTYStream()
        with mock.patch("lunar_tools.logprint._get_terminal_width", return_value=10):
            dynamic_print("Hello", stream=stream, persist=True)
        output = stream.getvalue()
        self.assertTrue(output.endswith("\n"))
        self.assertIn("Hello", output)

    def test_dynamic_print_non_tty_falls_back_to_plain_print(self):
        """dynamic_print should degrade gracefully when stream is not a TTY."""
        stream = io.StringIO()
        dynamic_print("Plain message", stream=stream)
        self.assertEqual(stream.getvalue(), "Plain message\n")

    def test_dynamic_print_truncates_long_messages(self):
        """Long messages should be truncated to fit the terminal width."""
        class FakeTTYStream(io.StringIO):
            def isatty(self):
                return True

        stream = FakeTTYStream()
        with mock.patch("lunar_tools.logprint._get_terminal_width", return_value=10):
            dynamic_print("This is a very long message", stream=stream)
        self.assertIn("This is...", stream.getvalue())

if __name__ == '__main__':
    unittest.main()



import time
import pytest
from lunar_tools.fps_tracker import FPSTracker

def test_segment_duration():
    # Set update_interval to 0 so update() always triggers display update.
    tracker = FPSTracker(update_interval=0)
    tracker.start_segment("test")
    # Sleep for 100ms
    time.sleep(0.1)
    # Call update to record the segment duration
    tracker.update()
    # Check that the segment "test" was recorded with duration near 100ms.
    assert "test" in tracker.segments, "Segment 'test' not found in segments."
    duration = tracker.segments["test"]
    # Allow a tolerance of 50ms to account for timing variability
    assert abs(duration - 100) < 50, f"Segment duration {duration} is not within expected range of 100ms."

def test_fps_calculation():
    tracker = FPSTracker(update_interval=0)
    # Simulate 5 frames with a consistent delay of 0.05 seconds between frames.
    for _ in range(5):
        time.sleep(0.05)
        tracker.update()
    fps = tracker.get_fps()
    # The expected fps should be around 20. Allow a tolerance.
    assert 18 <= fps <= 22, f"FPS value {fps} is not within the expected range (18-22 FPS)."

def test_segment_cleared_after_print():
    tracker = FPSTracker(update_interval=0)
    tracker.start_segment("seg1")
    time.sleep(0.05)
    tracker.update()
    # Call print_fps, which should print and clear segments if update_display occurred.
    tracker.print_fps()
    assert len(tracker.segments) == 0, "Segments were not cleared after printing FPS."

if __name__ == "__main__":
    pytest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests
from io import BytesIO
import time
import os
import tempfile
import base64
from PIL import Image
import numpy as np
from openai import OpenAI
import replicate
from lunar_tools.logprint import LogPrint
from lunar_tools.utils import read_api_key
import fal_client
from typing import Optional, Tuple


class FluxImageGenerator:
    ALLOWED_IMAGE_SIZES = [
        "square_hd", "square", "portrait_4_3", "portrait_16_9",
        "landscape_4_3", "landscape_16_9"
    ]

    def __init__(self, model="fal-ai/flux/schnell"):
        self.client = fal_client
        self.model = model
        self.last_result = None

    def generate(self, prompt: str, image_size: str = 'landscape_4_3', num_inference_steps: int = 4, seed: int = 420) -> Image.Image:
        if image_size not in self.ALLOWED_IMAGE_SIZES:
            raise ValueError(f"Invalid image size. Allowed sizes are: {', '.join(self.ALLOWED_IMAGE_SIZES)}")

        handler = self.client.submit(
            self.model,
            arguments={
                "prompt": prompt,
                "image_size": image_size,
                "num_inference_steps": num_inference_steps,
                "seed": seed
            },
        )

        result = handler.get()
        self.last_result = result
        image_url = result['images'][0]['url']

        # Download the image
        response = requests.get(image_url)
        response.raise_for_status()
        image = Image.open(BytesIO(response.content))
        return image


class NanoBananaEditImageGenerator:
    """
    Wrapper for fal.ai model: "fal-ai/nano-banana/edit".

    Notes:
    - Accepts a single image (base64 data URI or URL) via `image_url`, or a list via `image_urls`.
    - If a mask is provided, pass it via `mask_url` (base64 data URI or URL). Areas to edit should be white.
    - Additional tuning args supported by fal can be forwarded via **kwargs.
    - Returns a PIL.Image for the first image in the response.
    """

    def __init__(self, model="fal-ai/nano-banana/edit"):
        self.client = fal_client
        self.model = model
        self.last_result = None

    def generate(self,
                 prompt: str,
                 image_url=None,
                 image_urls=None,
                 mask_url=None,
                 seed=None,
                 num_images: int = 1,
                 sync_mode: bool = False,
                 image_size: str = None,
                 width: int = None,
                 height: int = None,
                 **kwargs):
        """
        Generate edited image(s) with Nano Banana edit.

        Args:
            prompt: Edit prompt.
            image_url: Single input image (URL or base64 data URI).
            image_urls: Multiple input images (list of URLs or base64 data URIs).
            mask_url: Optional mask image (URL or base64 data URI). White = edit.
            seed: Optional random seed for reproducibility.
            num_images: Number of images to generate (for single input image).
            sync_mode: If True, wait for image before returning.
            **kwargs: Forwarded to fal model arguments for advanced controls.
        """

        if not image_url and not image_urls:
            raise ValueError("Provide either image_url or image_urls")

        arguments = {
            "prompt": prompt,
            "sync_mode": sync_mode,
        }

        # Always send list field expected by the API
        if image_urls is not None:
            if not isinstance(image_urls, (list, tuple)):
                raise TypeError("image_urls must be a list of URLs/data URIs")
            arguments["image_urls"] = list(image_urls)
        else:
            arguments["image_urls"] = [image_url]

        if mask_url is not None:
            arguments["mask_url"] = mask_url
        if seed is not None:
            arguments["seed"] = seed
        if num_images is not None:
            arguments["num_images"] = int(num_images)
        if image_size is not None and image_size != "custom":
            arguments["image_size"] = image_size
        # If custom or explicit width/height provided, include them
        if width is not None:
            arguments["width"] = int(width)
        if height is not None:
            arguments["height"] = int(height)

        # Merge any additional kwargs (e.g., guidance_scale, steps, strength, negative_prompt, etc.)
        arguments.update(kwargs)

        # Verbose logging about the request
        try:
            num_inputs = len(arguments.get("image_urls", []))
        except Exception:
            num_inputs = 1 if arguments.get("image_url") else 0
        print(f"[NanoBanana] Submitting to {self.model} | images={num_inputs} | size={arguments.get('image_size') or 'custom'} | w={arguments.get('width')} | h={arguments.get('height')} | seed={arguments.get('seed')}")

        def _on_update(update):
            try:
                status = getattr(update, 'status', None) or update.get('status')
                message = getattr(update, 'message', None) or update.get('message')
                if status or message:
                    print(f"[NanoBanana][queue] status={status} msg={message}")
            except Exception:
                pass

        result = self.client.subscribe(
            self.model,
            arguments=arguments,
            with_logs=True,
            on_queue_update=_on_update
        )
        self.last_result = result

        # Extract first image URL from known result shapes
        image_url_out = None
        if isinstance(result, dict):
            if "images" in result and result["images"]:
                image_url_out = result["images"][0].get("url")
            elif "data" in result and isinstance(result["data"], dict):
                images = result["data"].get("images")
                if images:
                    image_url_out = images[0].get("url")

        if not image_url_out:
            raise ValueError("NanoBanana: could not find images in result")

        print(f"[NanoBanana] Received response. Downloading image: {image_url_out}")
        response = requests.get(image_url_out)
        response.raise_for_status()
        img = Image.open(BytesIO(response.content))
        print(f"[NanoBanana] Image downloaded. size={img.size} mode={img.mode}")
        return img


class FluxKontextImageGenerator:
    ALLOWED_ASPECT_RATIOS = [
        "21:9", "16:9", "4:3", "3:2", "1:1", "2:3", "3:4", "9:16", "9:21"
    ]
    
    ALLOWED_OUTPUT_FORMATS = ["jpeg", "png"]
    
    ALLOWED_SAFETY_TOLERANCE = ["1", "2", "3", "4", "5", "6"]

    def __init__(self, model="fal-ai/flux-pro/kontext"):
        self.client = fal_client
        self.model = model
        self.last_result = None

    def generate(self, prompt, image_url, seed=None, guidance_scale=3.5, num_images=1, 
                 output_format="jpeg", safety_tolerance="2", aspect_ratio=None, sync_mode=False):
        """
        Generate an edited image using Flux Kontext.
        
        Args:
            prompt (str): Description of what to change in the image
            image_url (str): URL or path to the input image
            seed (int, optional): Random seed for reproducible results
            guidance_scale (float): CFG scale (default: 3.5)
            num_images (int): Number of images to generate (default: 1)
            output_format (str): Output format - "jpeg" or "png" (default: "jpeg")
            safety_tolerance (str): Safety level 1-6, 1 most strict (default: "2")
            aspect_ratio (str, optional): Aspect ratio like "16:9", "1:1", etc.
            sync_mode (bool): Wait for image before returning (default: False)
        """
        if output_format not in self.ALLOWED_OUTPUT_FORMATS:
            raise ValueError(f"Invalid output format. Allowed formats are: {', '.join(self.ALLOWED_OUTPUT_FORMATS)}")
            
        if safety_tolerance not in self.ALLOWED_SAFETY_TOLERANCE:
            raise ValueError(f"Invalid safety tolerance. Allowed values are: {', '.join(self.ALLOWED_SAFETY_TOLERANCE)}")
            
        if aspect_ratio and aspect_ratio not in self.ALLOWED_ASPECT_RATIOS:
            raise ValueError(f"Invalid aspect ratio. Allowed ratios are: {', '.join(self.ALLOWED_ASPECT_RATIOS)}")

        arguments = {
            "prompt": prompt,
            "image_url": image_url,
            "guidance_scale": guidance_scale,
            "num_images": num_images,
            "output_format": output_format,
            "safety_tolerance": safety_tolerance,
            "sync_mode": sync_mode
        }
        
        # Add optional parameters if provided
        if seed is not None:
            arguments["seed"] = seed
        if aspect_ratio:
            arguments["aspect_ratio"] = aspect_ratio

        result = self.client.subscribe(
            self.model,
            arguments=arguments,
            with_logs=True,
            on_queue_update=lambda update: None  # You can add logging here if needed
        )
        self.last_result = result
        
        # Try to access the image URL correctly based on the result structure
        if 'images' in result:
            image_url = result['images'][0]['url']
        elif 'data' in result and 'images' in result['data']:
            image_url = result['data']['images'][0]['url']
        else:
            raise ValueError("Could not find images in result structure")

        # Download the image
        response = requests.get(image_url)
        response.raise_for_status()
        image = Image.open(BytesIO(response.content))
        return image


class Dalle3ImageGenerator:
    def __init__(self,
                 client=None,
                 logger=None,
                 model="dall-e-3",
                 size_output=(1792, 1024),
                 quality="standard"):
        if client is None:
            api_key = read_api_key("OPENAI_API_KEY")
            if not api_key:
                raise ValueError("No OPENAI_API_KEY found in environment variables")
            self.client = OpenAI(api_key=api_key)
        else:
            if not isinstance(client, OpenAI):
                raise TypeError("Invalid client type. Expected a 'OpenAI' instance.")
            self.client = client

        self.logger = logger if logger else LogPrint()
        self.model = model
        self.quality = quality
        self.set_dimensions(size_output)

    def set_dimensions(self, size_output):
        allowed_sizes = ["1024x1024", "1024x1792", "1792x1024"]

        # Check if size_output has a length of 2
        if len(size_output) != 2:
            raise ValueError("size_output must have a length of 2.")

        # Convert the input tuple to a string format
        size_str = f"{size_output[0]}x{size_output[1]}"

        if size_str not in allowed_sizes:
            raise ValueError("Invalid size. Allowed sizes are 1024x1024, 1024x1792, and 1792x1024.")
        else:
            self.size = size_str

    def generate(self, prompt: str, simulation: bool = False) -> Tuple[Optional[Image.Image], Optional[str]]:
        if simulation:
            # Simulation mode: Generate a random image
            width, height = map(int, self.size.split('x'))
            image_array = np.random.randint(0, 255, (height, width, 3), dtype=np.uint8)
            image = Image.fromarray(image_array, 'RGB')
            self.logger.print("Dalle3ImageGenerator: Simulation mode - random image generated")
            revised_prompt = "Simulation mode - no revised prompt"
        else:
            # Normal mode: Call the API to generate an image
            try:
                self.logger.print("Dalle3ImageGenerator: Starting image generation")
                start_time = time.time()

                response = self.client.images.generate(
                    model=self.model,
                    prompt=prompt,
                    size=self.size,
                    quality=self.quality,
                    n=1,
                )

                image_url = response.data[0].url
                response_http = requests.get(image_url)
                response_http.raise_for_status()

                image_data = BytesIO(response_http.content)
                image = Image.open(image_data)
                end_time = time.time()
                revised_prompt = response.data[0].revised_prompt
                self.logger.print(f"Dalle3ImageGenerator: Generation complete. Time taken: {int(end_time - start_time)} seconds")

            except requests.exceptions.RequestException as e:
                self.logger.print(f"HTTP Request failed: {e}")
                return None, None
            except Exception as e:
                self.logger.print(f"An error occurred: {e}")
                return None, None

        return image, revised_prompt


class SDXL_LCM:
    def __init__(self, client=None, logger=None, size_output=(1024, 1024), num_inference_steps=4):
        if client is None:
            self.client = replicate.Client(api_token=read_api_key("REPLICATE_API_TOKEN"))
        else:
            if not isinstance(client, replicate.Client):
                raise TypeError("Invalid client type. Expected a 'replicate.Client' instance.")
            self.client = client

        self.logger = logger if logger else LogPrint()
        self.size = size_output
        self.num_inference_steps = num_inference_steps

    def set_dimensions(self, size_output):
        self.size = size_output

    def set_num_inference_steps(self, num_inference_steps):
        self.num_inference_steps = num_inference_steps

    def generate(self, prompt, negative_prompt="", simulation=False):
        width, height = self.size
        num_inference_steps = self.num_inference_steps

        if simulation:
            # Simulation mode: Generate a random image
            image_array = np.random.randint(0, 255, (height, width, 3), dtype=np.uint8)
            image = Image.fromarray(image_array, 'RGB')
            self.logger.print("SDXL_LCM: Simulation mode - random image generated")
            img_url = "Simulation mode - no image URL"
            return image, img_url
        else:
            # Normal mode: Call the API to generate an image
            try:
                self.logger.print("SDXL_LCM: Starting image generation")
                start_time = time.time()

                output = self.client.run(
                    "lucataco/sdxl-lcm:fbbd475b1084de80c47c35bfe4ae64b964294aa7e237e6537eed938cfd24903d",
                    input={
                        "prompt": prompt,
                        "negative_prompt": negative_prompt,
                        "width": width,
                        "height": height,
                        "num_inference_steps": num_inference_steps
                    }
                )

                img_url = output[0]
                response_http = requests.get(img_url)
                response_http.raise_for_status()

                image_data = BytesIO(response_http.content)
                image = Image.open(image_data)
                end_time = time.time()
                return image, img_url
                self.logger.print(f"SDXL_LCM: Generation complete. Time taken: {int(end_time - start_time)} seconds")

            except requests.exceptions.RequestException as e:
                self.logger.print(f"HTTP Request failed: {e}")
                return None, None
           
class SDXL_TURBO:
    def __init__(self, client=None, logger=None, size_output=(512, 512), num_inference_steps=1):
        if client is None:
            self.client = replicate.Client(api_token=read_api_key("REPLICATE_API_TOKEN"))
        else:
            if not isinstance(client, replicate.Client):
                raise TypeError("Invalid client type. Expected a 'replicate.Client' instance.")
            self.client = client

        self.logger = logger if logger else LogPrint()
        self.size = size_output
        self.num_inference_steps = num_inference_steps

    def set_dimensions(self, size_output):
        self.size = size_output

    def set_num_inference_steps(self, num_inference_steps):
        self.num_inference_steps = num_inference_steps

    def generate(self, prompt, negative_prompt="", simulation=False):
        width, height = self.size
        num_inference_steps = self.num_inference_steps

        if simulation:
            # Simulation mode: Generate a random image
            image_array = np.random.randint(0, 255, (height, width, 3), dtype=np.uint8)
            image = Image.fromarray(image_array, 'RGB')
            self.logger.print("SDXL_TURBO: Simulation mode - random image generated")
            img_url = "Simulation mode - no image URL"
            return image, img_url
        else:
            # Normal mode: Call the API to generate an image
            try:
                self.logger.print("SDXL_TURBO: Starting image generation")
                start_time = time.time()

                output = self.client.run(
                    "fofr/sdxl-turbo:6244ebc4d96ffcc48fa1270d22a1f014addf79c41732fe205fb1ff638c409267",
                    input={
                        "prompt": prompt,
                        "agree_to_research_only": True,
                        "negative_prompt": negative_prompt,
                        "width": width,
                        "height": height,
                        "num_inference_steps": num_inference_steps
                    }
                )

                img_url = output[0]
                response_http = requests.get(img_url)
                response_http.raise_for_status()

                image_data = BytesIO(response_http.content)
                image = Image.open(image_data)
                end_time = time.time()
                return image, img_url
                self.logger.print(f"SDXL_TURBO: Generation complete. Time taken: {int(end_time - start_time)} seconds")

            except requests.exceptions.RequestException as e:
                self.logger.print(f"HTTP Request failed: {e}")
                return None, None


class GlifAPI:
    def __init__(self, api_token=None):
        if api_token is None:
            api_token = read_api_key("GLIF_API_KEY")
        self.base_url = "https://simple-api.glif.app"
        self.headers = {
            "Authorization": f"Bearer {api_token}",
            "Content-Type": "application/json"
        }

    def run_glif(self, glif_id, inputs, timeout=60):
        """Run a glif with given inputs, with error handling and timeout."""
        url = f"{self.base_url}/{glif_id}"
        payload = {"inputs": inputs}
        try:
            response = requests.post(url, json=payload, headers=self.headers, timeout=timeout)
            response.raise_for_status()  # Raises an HTTPError if the response status code is 4XX/5XX
            result = response.json()
            if 'output' in result and len(result['output']) > 0:
            
                image_url = result['output']
                if image_url.endswith(('.jpg', '.png')):
                    image_response = requests.get(image_url)
                    image_response.raise_for_status()
                    image = Image.open(BytesIO(image_response.content))
                    return {"image": image}
        except requests.exceptions.HTTPError as errh:
            return {"error": "Http Error", "message": str(errh)}
        except requests.exceptions.ConnectionError as errc:
            return {"error": "Connection Error", "message": str(errc)}
        except requests.exceptions.Timeout as errt:
            return {"error": "Timeout Error", "message": str(errt)}
        except requests.exceptions.RequestException as err:
            return {"error": "Something went wrong", "message": str(err)}


        return result

# Example usage
if __name__ == "__main__":

    flux = FluxImageGenerator()
    prompt_text = "photo of a person holding a sign with flüx written on it"
    image_size = "square_hd"
    num_inference_steps = 4
    seed = 420

    image = flux.generate(prompt_text, image_size, num_inference_steps, seed)

    
    # Example usage Flux Kontext - image to image editing
    print("Running Flux Kontext example...")
    
    # Check if we already have an original image
    if os.path.exists("original.jpg"):
        print("Using existing original.jpg")
    else:
        # Download a random image and save it
        temp_url = "https://picsum.photos/600/400"
        print(f"Downloading image from: {temp_url}")
        
        response = requests.get(temp_url)
        response.raise_for_status()
        original_image = Image.open(BytesIO(response.content))
        original_image.save("original.jpg")
        print("Original image saved as: original.jpg")
    
    # Convert the local file to base64 data URI
    print("Converting original.jpg to base64...")
    with open("original.jpg", "rb") as f:
        image_data = f.read()
        image_base64 = base64.b64encode(image_data).decode('utf-8')
        data_uri = f"data:image/jpeg;base64,{image_base64}"
    print("Image converted to base64 data URI")
    
    # Use Flux Kontext to edit the image
    flux_kontext = FluxKontextImageGenerator()
    edit_prompt = "Change the colors to sunset colors but keep the original shapes"
    
    # Generate edited image using the base64 data URI
    edited_image = flux_kontext.generate(
        prompt=edit_prompt,
        image_url=data_uri,
        seed=420,
        guidance_scale=3.5
    )
    
    print("Flux Kontext editing complete!")
    
    # Save the edited result in current directory
    edited_image.save("edited.jpg")
    print("Edited image saved as: edited.jpg")
    
    # Show the images if running interactively
    # original_image.show()
    # edited_image.show()

    
    # Example usage glifapi
    # glif = GlifAPI()
    # glif_id = "clgh1vxtu0011mo081dplq3xs"
    # inputs = {"node_6": "cute friendly oval shaped bot friend"}
    # result = glif.run_glif(glif_id, inputs)
    # print(result)


    # Example usage Dalle3
    # dalle3 = Dalle3ImageGenerator()
    # image, revised_prompt = dalle3.generate("realistic photo of a ")
    # image = image.resize((1024, 576))
    # image.save("/Users/jjj/glif/git_remote/generative-models/assets/fluid3.jpg")
    # image.show()

    # output = replicate.run(
    #     "lucataco/sdxl-lcm:fbbd475b1084de80c47c35bfe4ae64b964294aa7e237e6537eed938cfd24903d",
    #     input={"prompt": "An astronaut riding a rainbow unicorn, cinematic, dramatic",
    #            "negative_prompt": "cartoon",
    #            "width": 1280,
    #            "height": 1024,
    #            "num_inference_steps": 6}
    # )
    # img_url = output[0]
    
    # Example usage
    # client = OpenAI()
    # sdxl_turbo = SDXL_TURBO()
    # image, img_url = sdxl_turbo.generate("An astronaut riding a rainbow unicorn", "cartoon")

    # sdxl_lcm = SDXL_LCM()
    # image, img_url = sdxl_lcm.generate("An astronaut riding a rainbow unicorn", "cartoon")


    



# based on https://github.com/jbaron34/torchwindow

import sys
import ctypes
import warnings
import numpy as np
try:
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False
    torch = None
from sys import platform
from PIL import Image
import cv2
import random
from lunar_tools.utils import get_os_type
import pygame
import threading
import time

if get_os_type() == "Linux":
    try:
        from cuda import cudart as cu
    except ImportError:
        try:
            from cuda.bindings import runtime as cu
        except ImportError:
            raise ImportError("Could not import CUDA runtime. Please ensure cuda-python is properly installed.")
    
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=UserWarning)
        import sdl2
        import sdl2.ext
    
    from sdl2 import video
    from OpenGL import GL as gl   

import logging

logger = logging.getLogger(__name__)

# CUDA/SDL Exception classes
class SDLException(Exception):
    pass

class CudaException(Exception):
    pass

class OpenGLException(Exception):
    pass

# Shaders
from OpenGL.GL import GL_VERTEX_SHADER, GL_FRAGMENT_SHADER
from OpenGL.GL.shaders import compileProgram, compileShader


VERTEX_SHADER_SOURCE = """
#version 450

smooth out vec4 fragColor;
smooth out vec2 texcoords;

vec4 positions[3] = vec4[3](
    vec4(-1.0, 1.0, 0.0, 1.0),
    vec4(3.0, 1.0, 0.0, 1.0),
    vec4(-1.0, -3.0, 0.0, 1.0)
);

vec2 texpos[3] = vec2[3](
    vec2(0, 0),
    vec2(2, 0),
    vec2(0, 2)
);

vec4 colors[3] = vec4[3](
    vec4(1.0, 0.0, 0.0, 1.0),
    vec4(0.0, 1.0, 0.0, 1.0),
    vec4(0.0, 0.0, 1.0, 1.0)
);

void main() {
    gl_Position = positions[gl_VertexID];
    fragColor = colors[gl_VertexID];
    texcoords = texpos[gl_VertexID];
}
"""

FRAGMENT_SHADER_SOURCE = """
#version 330

smooth in vec4 fragColor;
smooth in vec2 texcoords;

out vec4 outputColour;

uniform sampler2D texSampler;

void main()
{
    outputColour = texture(texSampler, texcoords);
}
"""


def create_shader_program():
    vertex_shader = compileShader(VERTEX_SHADER_SOURCE, GL_VERTEX_SHADER)
    fragment_shader = compileShader(FRAGMENT_SHADER_SOURCE, GL_FRAGMENT_SHADER)
    return compileProgram(vertex_shader, fragment_shader)

class PeripheralEvent():
    def __init__(self):
        self.keycode = -1
        self.mouse_button_state = -1
        self.mouse_posX = -1
        self.mouse_posY = -1

# keycode conversion function  SDL2 -> OpenCV convention
def sdl_to_cv2_keycode(sdl_keycode):
    # Map for common keys (alphabets and numbers)
    if sdl2.SDL_SCANCODE_A <= sdl_keycode <= sdl2.SDL_SCANCODE_Z:
        return ord(chr(sdl_keycode - sdl2.SDL_SCANCODE_A + ord('a')))
    elif sdl2.SDL_SCANCODE_1 <= sdl_keycode <= sdl2.SDL_SCANCODE_0:
        return ord(chr(sdl_keycode - sdl2.SDL_SCANCODE_1 + ord('1')))
    
    # Digits
        if sdl2.SDL_SCANCODE_1 <= sdl_keycode <= sdl2.SDL_SCANCODE_0:
            return ord(chr(sdl_keycode - sdl2.SDL_SCANCODE_1 + ord('1')))    
    
    special_keys_map = {
        sdl2.SDL_SCANCODE_RETURN: 13,     # Enter key
        sdl2.SDL_SCANCODE_ESCAPE: 27,     # Escape key
        sdl2.SDL_SCANCODE_BACKSPACE: 8,   # Backspace key
        sdl2.SDL_SCANCODE_TAB: 9,         # Tab key
        sdl2.SDL_SCANCODE_SPACE: 32,      # Space key
        sdl2.SDL_SCANCODE_F1: 0x700000,   # F1 key
        sdl2.SDL_SCANCODE_F2: 0x710000,   # F2 key
        sdl2.SDL_SCANCODE_RETURN: 13,     # Enter key
        sdl2.SDL_SCANCODE_ESCAPE: 27,     # Escape key
        sdl2.SDL_SCANCODE_BACKSPACE: 8,   # Backspace key
        sdl2.SDL_SCANCODE_TAB: 9,         # Tab key
        sdl2.SDL_SCANCODE_SPACE: 32,      # Space key
        sdl2.SDL_SCANCODE_F1: 0x700000,   # F1 key
        sdl2.SDL_SCANCODE_F2: 0x710000,   # F2 key
        sdl2.SDL_SCANCODE_F3: 0x720000,   # F3 key
        sdl2.SDL_SCANCODE_F4: 0x730000,   # F4 key
        # ... (Continue for other function keys F5 to F12)
        sdl2.SDL_SCANCODE_RIGHT: 0x270000,  # Right arrow key
        sdl2.SDL_SCANCODE_LEFT: 0x250000,   # Left arrow key
        sdl2.SDL_SCANCODE_DOWN: 0x280000,   # Down arrow key
        sdl2.SDL_SCANCODE_UP: 0x260000,     # Up arrow key        
        # Add other special keys here
        # ...
    }
    
    if sdl_keycode in special_keys_map:
        return special_keys_map.get(sdl_keycode, -1)
    elif sdl_keycode == -1:
        pass
    else:
        print('sdl_to_cv2_keycode -> unknown key code')
        return -1

class Renderer:
    def __init__(self, width: int = 1920, height: int = 1080, 
                 gpu_id: int = 0,
                 window_title: str = "lunar_render_window",
                 do_fullscreen: bool = False,
                 do_window_refresh_no_freeze = True,  # TODO: setting it to True might interfere with event polling, such as grabbing inputs from keyboard/mouse; so far only implemented for "gl" backend
                 backend = None,
                 display_id: int = None):
        
        self.window_title = window_title
        self.gpu_id = gpu_id
        self.width = width
        self.height = height
        self.do_fullscreen = do_fullscreen
        self.do_window_refresh_no_freeze = do_window_refresh_no_freeze
        self.display_id = display_id
        
        if backend is None:
            if get_os_type() == "Linux":
                self.backend = 'gl'
            else:
                self.backend = 'opencv'
        else:
            assert backend in ['gl', 'opencv', 'pygame', 'cudagl']
		
            if get_os_type() == "MacOS":
              if backend != 'opencv' and backend != 'pygame':
                backend = 'pygame'

            self.backend = backend
        
        if self.backend == 'cudagl':
            self.cuda_is_setup = False
            self.cudasdl_setup()
            self.cudagl_setup()
            self.cuda_setup()
            self.running = True
        elif self.backend == 'gl':
            self.sdl_setup()
            self.running = True
        elif self.backend == 'pygame':
            self.pygame_setup(self.display_id)
            
    def sdl_setup(self):
        # Initialize SDL2
        sdl2.ext.init()
        
        if self.do_fullscreen:
            flags_window = sdl2.SDL_WINDOW_OPENGL | sdl2.SDL_WINDOW_FULLSCREEN
        else:
            flags_window = sdl2.SDL_WINDOW_OPENGL
        
        # Create an SDL2 window
        self.sdl_window = sdl2.SDL_CreateWindow(b"Image Viewer", 
                                            sdl2.SDL_WINDOWPOS_CENTERED, 
                                            sdl2.SDL_WINDOWPOS_CENTERED, 
                                            self.width, self.height, 
                                            flags_window)
        if not self.sdl_window:
            raise Exception("Could not create SDL window:", sdl2.SDL_GetError())
        
        # Create OpenGL context
        self.context = sdl2.SDL_GL_CreateContext(self.sdl_window)
        
        # Enable 2D textures
        gl.glEnable(gl.GL_TEXTURE_2D)
        
        # Generate a texture ID and bind it
        self.tex = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.tex)
        
        # Set texture parameters
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        
        # Event polling thread
        if self.do_window_refresh_no_freeze:
            self.running = True
            self.event_thread = threading.Thread(target=self.gl_poll_events_thread, daemon=True)
            self.event_thread.start()

    def cudasdl_setup(self):
        if sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO):
            raise SDLException(sdl2.SDL_GetError())

        if self.do_fullscreen:
            flags_window = sdl2.SDL_WINDOW_OPENGL | sdl2.SDL_WINDOW_FULLSCREEN
        else:
            flags_window = sdl2.SDL_WINDOW_OPENGL

        self.sdl_window = sdl2.SDL_CreateWindow(
            self.window_title.encode(),
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            self.width,
            self.height,
            flags_window,
        )
        
        if not self.sdl_window:
            raise SDLException(sdl2.SDL_GetError())

        # Force OpenGL 3.3 'core' context.
        # Must set *before* creating GL context!
        video.SDL_GL_SetAttribute(video.SDL_GL_CONTEXT_MAJOR_VERSION, 3)
        video.SDL_GL_SetAttribute(video.SDL_GL_CONTEXT_MINOR_VERSION, 3)
        video.SDL_GL_SetAttribute(
            video.SDL_GL_CONTEXT_PROFILE_MASK, video.SDL_GL_CONTEXT_PROFILE_CORE
        )
        self.gl_context = sdl2.SDL_GL_CreateContext(self.sdl_window)
        
    def cudagl_setup(self):
        self.shader_program = create_shader_program()
        self.vao = gl.glGenVertexArrays(1)

        self.tex = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.tex)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        gl.glTexImage2D(
            gl.GL_TEXTURE_2D,
            0,
            gl.GL_RGBA32F,
            self.width,
            self.height,
            0,
            gl.GL_RGBA,
            gl.GL_FLOAT,
            None,
        )
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

    def cuda_setup(self):
        if self.cuda_is_setup:
            return

        if sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO) != 0:
            raise SDLException(sdl2.SDL_GetError())

        err, *_ = cu.cudaGLGetDevices(1, cu.cudaGLDeviceList.cudaGLDeviceListAll)
        if err == cu.cudaError_t.cudaErrorUnknown:
            raise OpenGLException(
                "OpenGL context may be running on integrated graphics"
            )

        err, self.cuda_image = cu.cudaGraphicsGLRegisterImage(
            self.tex,
            gl.GL_TEXTURE_2D,
            cu.cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsWriteDiscard,
        )
        if err != cu.cudaError_t.cudaSuccess:
            raise CudaException("Unable to register opengl texture")

        self.cuda_is_setup = True

    def cudagl_draw_internal(self):
        gl.glUseProgram(self.shader_program)
        try:
            gl.glClearColor(0, 0, 0, 1)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)
            gl.glBindTexture(gl.GL_TEXTURE_2D, self.tex)
            gl.glBindVertexArray(self.vao)
            gl.glDrawArrays(gl.GL_TRIANGLES, 0, 3)
        finally:
            gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
            gl.glBindVertexArray(0)
            gl.glUseProgram(0)
        sdl2.SDL_GL_SwapWindow(self.sdl_window)

    def gl_get_events(self):
        event = sdl2.SDL_Event()
        
        # retrieve presses keyboard key codes
        key_states = sdl2.SDL_GetKeyboardState(None)
        
        # handle mouse presses
        mouse_posX, mouse_posY = ctypes.c_int(0), ctypes.c_int(0)

        pressed_key_code = -1
        mouse_buttonstate = -1
        if self.running:
            key_press_tracker_array = np.zeros((sdl2.SDL_NUM_SCANCODES,), dtype=np.int8)
            while sdl2.SDL_PollEvent(ctypes.byref(event)):
                mouse_buttonstate = sdl2.mouse.SDL_GetMouseState(ctypes.byref(mouse_posX), ctypes.byref(mouse_posY))
                
                if (event.type == sdl2.SDL_WINDOWEVENT and event.window.event == sdl2.SDL_WINDOWEVENT_CLOSE):
                    self.running = False
                    self.gl_close()
    
                # Exit code
                if key_states[sdl2.SDL_SCANCODE_ESCAPE]:
                    self.running = False
                    self.gl_close()
                    sys.exit(0)
                
                for key_code in range(sdl2.SDL_NUM_SCANCODES):
                    # key pressed
                    if key_states[key_code] and key_press_tracker_array[key_code] == 0:
                        key_press_tracker_array[key_code] == 1
                        pressed_key_code = key_code
                        
                    # key released
                    elif not key_states[key_code] and key_press_tracker_array[key_code] == 1:
                        key_press_tracker_array[key_code] == 0
            
            if self.backend == 'cudagl':
                self.cudagl_draw_internal()
            
        # convert key code to SDL2 convention
        pressed_key_code = sdl_to_cv2_keycode(pressed_key_code)
        
        peripheralEvent = PeripheralEvent()
        peripheralEvent.pressed_key_code = pressed_key_code
        peripheralEvent.mouse_button_state = mouse_buttonstate
        peripheralEvent.mouse_posX = mouse_posX.value
        peripheralEvent.mouse_posY = mouse_posY.value
        
        return peripheralEvent
    
    def gl_render(self, image):

        if TORCH_AVAILABLE and isinstance(image, torch.Tensor):
            image = image.cpu().numpy()
        else:
            if type(image) == np.ndarray:
                pass
            elif type(image) == Image.Image:
                image = np.array(image)
            else:
                raise Exception('render function received input of unknown type')        
                
        # clamp and set correct type
        image = np.clip(image, 0, 255)
        image = image.astype(np.uint8)
        
        # reshape if there is a mismatched between supply image size and window size
        if image.shape[1] != self.width or image.shape[0] != self.height:
            image = cv2.resize(image, (self.width, self.height))
            
        image = np.flip(image, axis=0).copy()
            
        # Load image data into the texture
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.tex)
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB, self.width, self.height, 0, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, image)
        
        events = sdl2.ext.get_events()

        # Handle SDL2 events
        # events = sdl2.ext.get_events()
        # for event in events:
        #     if event.type == sdl2.SDL_QUIT:
        #         self.cleanup()
        #         return

        # Clear the screen
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        
        # Render the texture as a quad
        gl.glBegin(gl.GL_QUADS)
        gl.glTexCoord2f(0.0, 0.0)
        gl.glVertex2f(-1.0, -1.0)
        gl.glTexCoord2f(1.0, 0.0)
        gl.glVertex2f(1.0, -1.0)
        gl.glTexCoord2f(1.0, 1.0)
        gl.glVertex2f(1.0, 1.0)
        gl.glTexCoord2f(0.0, 1.0)
        gl.glVertex2f(-1.0, 1.0)
        gl.glEnd()
        
        # Swap buffers
        sdl2.SDL_GL_SwapWindow(self.sdl_window)
        
        pressed_key_code = self.gl_get_events()
        return pressed_key_code        

    def cudagl_render(self, image):
        if not TORCH_AVAILABLE:
            raise ImportError(
                "cudagl_render requires PyTorch. Install it via 'pip install torch' "
                "or 'pip install lunar-tools[torch]'."
            )

        # first check if input data types are valid
        if isinstance(image, torch.Tensor):
            if image.device.type == 'cpu':
                # force placement on GPU
                image = image.to(f'cuda:{self.gpu_id}')
        else:
            # cast as torch tensor / place on GPU
            if type(image) == np.ndarray:
                image = torch.from_numpy(image).to(f'cuda:{self.gpu_id}')
            elif type(image) == Image.Image:
                image = torch.from_numpy(np.array(image)).to(f'cuda:{self.gpu_id}')
            else:
                raise Exception('render function received input of unknown type')
                
        # bring to OpenGL-standard range
        image = image.float() / 255
                
        # clamp to the valid range
        image = torch.clamp(image, 0, 1)
        
        # resize 
        if image.shape[1] != self.width or image.shape[0] != self.height:
            image = torch.nn.functional.interpolate(image.permute([2,0,1])[None], size=(self.height, self.width))
            image = image[0].permute([1,2,0])
        
        # transpose X/Y for openGL consistency
        # image = image.permute((1,0,2))
                
        # check for number of channels
        if len(image.shape) == 3:
            if image.shape[2] == 3: # add rgbA channel
                ones = torch.ones(image.size(0), image.size(1), 1, device=image.device)
                image = torch.cat((image, ones), -1)
            elif image.shape[2] == 4: # correct number of channels -> pass
                pass
            else:
                raise Exception('render function received the wrong number of channels')
        elif len(image.shape) == 2:  # grayscale input  -> to RGBA
            image = torch.cat((image, )*4, -1)
        else:
            raise Exception('render function received the wrong number of channels')
            
        # do rendering
        if not self.running:
            return
        if not self.cuda_is_setup:
            self.cuda_setup()
        (err,) = cu.cudaGraphicsMapResources(1, self.cuda_image, cu.cudaStreamLegacy)
        if err != cu.cudaError_t.cudaSuccess:
            raise CudaException("Unable to map graphics resource")
        err, array = cu.cudaGraphicsSubResourceGetMappedArray(self.cuda_image, 0, 0)
        if err != cu.cudaError_t.cudaSuccess:
            raise CudaException("Unable to get mapped array")
        (err,) = cu.cudaMemcpy2DToArrayAsync(
            array,
            0,
            0,
            image.data_ptr(),
            4 * 4 * self.width,
            4 * 4 * self.width,
            self.height,
            cu.cudaMemcpyKind.cudaMemcpyDeviceToDevice,
            cu.cudaStreamLegacy,
        )
        if err != cu.cudaError_t.cudaSuccess:
            raise CudaException("Unable to copy from tensor to texture")

        (err,) = cu.cudaGraphicsUnmapResources(1, self.cuda_image, cu.cudaStreamLegacy)
        if err != cu.cudaError_t.cudaSuccess:
            raise CudaException("Unable to unmap graphics resource")
        pressed_key_code = self.gl_get_events()
        return pressed_key_code
        
    def cv2_render(self, image):
        if TORCH_AVAILABLE and isinstance(image, torch.Tensor):
            image = image.cpu().numpy()
        else:
            if type(image) == np.ndarray:
                pass
            elif type(image) == Image.Image:
                image = np.array(image)
            else:
                raise Exception('render function received input of unknown type')        
                
        # clamp and set correct type
        image = np.clip(image, 0, 255)
        image = image.astype(np.uint8)
        
        # reshape if there is a mismatched between supply image size and window size
        if image.shape[1] != self.width or image.shape[0] != self.height:
            image = cv2.resize(image, (self.width, self.height))
        
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        
        cv2.imshow(self.window_title, image) 
        cv2_keycode = cv2.waitKey(1)
        if cv2_keycode == 27:  # exit code
            cv2.destroyAllWindows()
            sys.exit(0)
            
        if self.do_fullscreen:
            cv2.namedWindow(self.window_title,cv2.WND_PROP_FULLSCREEN)
            cv2.setWindowProperty(self.window_title, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
            
        peripheralEvent = PeripheralEvent()
        peripheralEvent.pressed_key_code = cv2_keycode
        
        # not implemented
        peripheralEvent.mouse_button_state = -1
        peripheralEvent.mouse_posX = -1
        peripheralEvent.mouse_posY = -1            
            
        return peripheralEvent

    def pygame_setup(self, display_id=None):
        pygame.init()
        flags = (pygame.NOFRAME | pygame.RESIZABLE) if self.do_fullscreen else 0
        self.screen = pygame.display.set_mode((self.width, self.height), flags, display=display_id)
        pygame.display.set_caption(self.window_title)
        self.running = True

    def pygame_render(self, image):
        if TORCH_AVAILABLE and isinstance(image, torch.Tensor):
            image = image.cpu().numpy()
        else:
            if type(image) == np.ndarray:
                pass
            elif type(image) == Image.Image:
                image = np.array(image)
            else:
                raise Exception('render function received input of unknown type')
        
        # clamp and set correct type
        image = np.clip(image, 0, 255).astype(np.uint8)

        # reshape if there is a mismatched between supply image size and window size
        if image.shape[1] != self.width or image.shape[0] != self.height:
            image = cv2.resize(image, (self.width, self.height))

        # image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        image = np.rot90(image)  # Pygame requires rotation to match orientation
        image = np.flip(image, axis=0)

        surf = pygame.surfarray.make_surface(image)
        self.screen.blit(surf, (0, 0))
        pygame.display.flip()

        pressed_key_code = None
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
                pygame.quit()
                sys.exit(0)
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    self.running = False
                    pygame.quit()
                    sys.exit(0)
                pressed_key_code = event.key

        peripheralEvent = PeripheralEvent()
        peripheralEvent.pressed_key_code = pressed_key_code
        
        # not implemented
        peripheralEvent.mouse_button_state = -1
        peripheralEvent.mouse_posX = -1
        peripheralEvent.mouse_posY = -1
            
        return peripheralEvent
            
    def render(self, image):
        if self.backend == 'cudagl':
            peripheralEvent = self.cudagl_render(image)
        elif self.backend == 'gl':
            peripheralEvent = self.gl_render(image)
        elif self.backend == 'opencv':
            peripheralEvent = self.cv2_render(image)
        elif self.backend == 'pygame':
            peripheralEvent = self.pygame_render(image)
        return peripheralEvent

    def gl_close(self):
        self.running = False
        sdl2.SDL_GL_DeleteContext(self.gl_context)
        sdl2.SDL_DestroyWindow(self.sdl_window)
        sdl2.SDL_Quit()
        
    def gl_poll_events_thread(self):
        # Poll SDL events in a separate thread
        while self.running:
            events = sdl2.ext.get_events()
            for event in events:
                if event.type == sdl2.SDL_QUIT:
                    self.cleanup()
                    exit(0)
            time.sleep(1.0)  # Prevent busy-looping           


class GridRenderer():
    def __init__(self, nmb_rows, nmb_cols, shape_hw, backend=None, window_title='lunar_render_window'):
        """
        nmb_rows: Number of tiles in vertical direction
        nmb_cols: Number of tiles in horizontal direction
        shape_hw: (H,W) = tuple (height,width) of individual tile
        """
        self.H = shape_hw[0]
        self.W = shape_hw[1]
        self.nmb_rows = nmb_rows
        self.nmb_cols = nmb_cols
        self.canvas = np.zeros((nmb_rows, nmb_cols, shape_hw[0], shape_hw[1], 3), dtype=np.uint8)
        self.renderer = Renderer(width=shape_hw[1]*nmb_cols, height=shape_hw[0]*nmb_rows, backend=backend, window_title=window_title)
        
    def update(self, tiles):
        """
        Concatenate image tiles into one large canvas.
    
        :param tiles: NumPy array of shape nmb_rowsxnmb_colsxHxW*C
                       nmb_rows: Number of tiles in vertical direction
                       nmb_cols: Number of tiles in horizontal direction
                       H: Height of each tile
                       W: Width of each tile
                       C: Number of RGB channels
        :return: NumPy array representing the large canvas with shape (nmb_rows*H)x(nmb_cols*W)
        """
        if isinstance(tiles, list):
            tiles = self.list_to_tiles(tiles)
        nmb_rows, nmb_cols, H, W, C = tiles.shape
        fail_msg = 'GridRenderer->inject_tiles: tiles shape inconsistent with initialization'
        assert (nmb_rows == self.nmb_rows) and (nmb_cols == self.nmb_cols), print(fail_msg)
        assert (H == self.H) and (W == self.W), print(fail_msg)
        
        # Reshape and transpose to bring tiles next to each other
        self.canvas = tiles.transpose(0, 2, 1, 3, 4).reshape(nmb_rows*H, nmb_cols*W, C)
        
    def list_to_tiles(self, list_images):
        """
        Reshape image tiles from list to tensor.
    
        :param list_images: list of images of shape H*W*3
        :return: NumPy array of shape nmb_rowsxnmb_colsxHxW*C
        """        
        
        grid_input = np.zeros((self.nmb_rows, self.nmb_cols, self.H, self.W, 3), dtype=np.uint8)
        for m in range(self.nmb_rows):
            for n in range(self.nmb_cols):
                if m*self.nmb_cols + n < len(list_images):
                    grid_input[m,n,:,:,:] = list_images[m*self.nmb_cols + n]
        return grid_input        

    def render(self):
        """
        Render canvas abd find the index of the tile given a mouse click pixel coordinate on the 2D canvas.
        :return: A tuple (m, n) representing the tile index in the range (0..nmb_rows, 0..nmb_cols).
        """
        
        peripheralEvent = self.renderer.render(self.canvas)
        
        if peripheralEvent.mouse_button_state > 0:
            x = peripheralEvent.mouse_posX
            y = peripheralEvent.mouse_posY
            
            m = y // self.H
            n = x // self.W
            return m, n
        else:
            return -1, -1
        
if __name__ == '__main__':
    import time
    nmb_rows = 2
    nmb_cols = 2
    shape_hw = (200, 300)
    gridrenderer = GridRenderer(nmb_rows, nmb_cols, shape_hw)


    while True:
        time.sleep(0.1)
        list_imgs = []
        for _ in range(nmb_rows * nmb_cols):
            img = Image.new('RGB', shape_hw[::-1], (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)))
            list_imgs.append(img)
        gridrenderer.update(list_imgs)
        gridrenderer.render()



if __name__ == '__main__':
    
    sz = (512, 1024)
    renderer = Renderer(width=sz[1], height=sz[0], backend='gl', do_fullscreen=False)
    
    while True:
        # numpy array
        image = np.random.rand(sz[0]//4,sz[1]//4,3)*255
        image[:50,:,:] = 0
        image[:,:50,:] = 0
        
        # PIL array
        # image = Image.fromarray(np.uint8(np.random.rand(sz[0],sz[1],4)*255))
        
        # Torch tensors
        # image = torch.rand((sz[0]//2,sz[1]//2,4))*255
        # image = torch.rand((sz[0],sz[1]), device='cuda:0')*255
        # image = torch.rand((sz[0],sz[1],3), device='cuda:0')*255
        # image = torch.rand((sz[0],sz[1],4), device='cuda:0')*255
        renderer.render(image)


    


import asyncio
import time
import threading

from lunar_tools import realtime_voice

def test_on_ai_audio_complete_existence():
    # Verify that the module does NOT expose 'onAudioComplete'
    assert not hasattr(realtime_voice, "onAudioComplete"), "Module should not have onAudioComplete"
    # Verify that the RealTimeVoice class exposes the renamed 'onAIAudioComplete'
    assert hasattr(realtime_voice.RealTimeVoice, "onAIAudioComplete"), "RealTimeVoice should have onAIAudioComplete"

def test_on_ai_audio_complete_functionality():
    flag = {"called": False}

    async def dummy_callback():
        flag["called"] = True

    # Create an instance of RealTimeVoice with the dummy callback.
    rtv = realtime_voice.RealTimeVoice(
        instructions="Test instructions",
        on_ai_audio_complete=dummy_callback,
        verbose=True
    )
    # Simulate that audio has been produced.
    rtv._audio_complete_pending = True

    # Set up an event loop for the instance.
    loop = asyncio.new_event_loop()
    rtv._loop = loop
    # Start the loop in a separate thread.
    def run_loop():
        loop.run_forever()
    t = threading.Thread(target=run_loop, daemon=True)
    t.start()

    # Call the onAIAudioComplete method to trigger the callback.
    rtv.onAIAudioComplete()
    # Allow a short time for the scheduled callback to run.
    time.sleep(0.2)
    # Clean up the loop.
    loop.call_soon_threadsafe(loop.stop)
    t.join(timeout=1)
    loop.close()

    assert flag["called"] == True, "The on_ai_audio_complete callback should have been called."



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from unittest.mock import patch, MagicMock
from lunar_tools.image_gen import Dalle3ImageGenerator, SDXL_LCM
from lunar_tools.utils import read_api_key
from PIL import Image
from openai import OpenAI
import replicate

# class TestDalle3ImageGenerator(unittest.TestCase):
#     def test_openai_client_initialization(self):

#         # Set up a mock environment variable
#         with patch.dict(os.environ, {'OPENAI_API_KEY': 'test_key'}):
#             generator = Dalle3ImageGenerator()
#             self.assertIsNotNone(generator.client)

#     def test_image_generation(self):
#         # Mock the generate method
#         size_output = (1024, 1024)

#         # Test image generation with specified size
#         generator = Dalle3ImageGenerator(size_output=size_output)
#         image, _ = generator.generate("a beautiful blue house")
#         self.assertIsNotNone(image)
#         self.assertEqual(image.size, size_output)
        
#     def test_correct_client(self):
#         """Test Dalle3ImageGenerator with the correct client type."""
#         client = OpenAI(api_key=read_api_key("OPENAI_API_KEY"))
#         try:
#             generator = Dalle3ImageGenerator(client=client)
#         except Exception as e:
#             self.fail(f"Instantiation with correct client type raised an exception: {e}")

#     def test_incorrect_client(self):
#         """Test Dalle3ImageGenerator with an incorrect client type."""
#         client = WrongClient()
#         with self.assertRaises(TypeError):
#             generator = Dalle3ImageGenerator(client=client)


class TestLCMSDXL(unittest.TestCase):

    def setUp(self):
        self.sdxl_lcm = SDXL_LCM()
        
    def test_correct_client(self):
        """Test the SDXL_LCM class with the correct client type."""
        client = replicate.Client(api_token=read_api_key("REPLICATE_API_TOKEN"))
        try:
            sdxl_lcm = SDXL_LCM(client=client)
        except Exception as e:
            self.fail(f"Instantiation with correct client type raised an exception: {e}")

    def test_incorrect_client(self):
        """Test the SDXL_LCM class with an incorrect client type."""
        client = OpenAI(api_key=read_api_key("OPENAI_API_KEY"))
        with self.assertRaises(TypeError):
            sdxl_lcm = SDXL_LCM(client=client)


    def test_generate_simulation_mode(self):
        # Test the generate function in simulation mode
        image, img_url = self.sdxl_lcm.generate("An astronaut riding a rainbow unicorn", simulation=True)

        # Assertions to check if the output is as expected
        self.assertIsNotNone(image)
        self.assertEqual(img_url, "Simulation mode - no image URL")



class WrongClient:
    def __init__(self):
        pass




if __name__ == '__main__':
    unittest.main()



import importlib.util
import os
import socket
import string
import sys
import threading
import time
import types
import unittest
from pathlib import Path

import numpy as np

if "elevenlabs.client" not in sys.modules:
    eleven_mod = types.ModuleType("elevenlabs")
    eleven_client_mod = types.ModuleType("elevenlabs.client")

    class _DummyElevenLabs:
        pass

    eleven_client_mod.ElevenLabs = _DummyElevenLabs
    eleven_mod.Voice = type("Voice", (), {})
    eleven_mod.VoiceSettings = type("VoiceSettings", (), {})

    def _noop(*_, **__):
        return None

    eleven_mod.play = _noop
    eleven_mod.save = _noop
    sys.modules["elevenlabs"] = eleven_mod
    sys.modules["elevenlabs.client"] = eleven_client_mod
    eleven_mod.client = eleven_client_mod

OSC_MODULE = Path(__file__).resolve().parents[1] / "lunar_tools" / "comms" / "osc.py"
spec = importlib.util.spec_from_file_location("lt_osc", OSC_MODULE)
osc_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(osc_module)
OSCSender = osc_module.OSCSender
OSCReceiver = osc_module.OSCReceiver


class TestOSCReceiverClient(unittest.TestCase):
    def setUp(self):
        port = self._reserve_port()
        if port is None:
            self.skipTest("UDP sockets are not permitted in this environment.")
        self.port = port
        self.sender = OSCSender('127.0.0.1', port_receiver=self.port)
        self.receiver = OSCReceiver('127.0.0.1', port_receiver=self.port)

    def tearDown(self):
        if hasattr(self, "receiver"):
            self.receiver.stop()

    @staticmethod
    def _reserve_port():
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
                sock.bind(("127.0.0.1", 0))
                return sock.getsockname()[1]
        except PermissionError:
            return None

    def test_received_short(self):
        time.sleep(0.1)
        self.sender.send_message("/env1", 42)
        time.sleep(0.1)
        data = self.receiver.get_all_values("/env1")
        if data is not None:
            self.assertEqual(len(data), 1)
            self.assertEqual(data[0], 42)
        else:
            self.fail("No data received from OSC receiver.")

    def test_received_long(self):
        time.sleep(0.1)
        for i in range(10):
            self.sender.send_message("/env1", i)
            time.sleep(0.1)
        data = self.receiver.get_all_values("/env1")
        if data is not None:
            self.assertEqual(len(data), 10)
            for i in range(10):
                self.assertEqual(data[i], i)
        else:
            self.fail("No data received from OSC receiver.")


class TestOSCReceiverUnit(unittest.TestCase):
    def setUp(self):
        self.receiver = OSCReceiver('127.0.0.1', start=False)

    def test_sender_requires_ip(self):
        with self.assertRaises(ValueError):
            OSCSender(ip_receiver=None)

    def test_process_incoming_requires_payload(self):
        self.receiver.process_incoming("/env")
        self.assertNotIn("/env", self.receiver.dict_messages)

    def test_process_incoming_non_numeric(self):
        self.receiver.process_incoming("/env", "not-a-number")
        self.assertNotIn("/env", self.receiver.dict_messages)

    def test_process_incoming_numeric(self):
        self.receiver.process_incoming("/env", 10)
        self.assertIn("/env", self.receiver.dict_messages)
        self.assertEqual(self.receiver.dict_messages["/env"][-1], 10.0)

    def test_stop_shuts_down_visualization(self):
        class DummyRenderer:
            def __init__(self):
                self.closed = False

            def close(self):
                self.closed = True

        self.receiver.running_vis = True
        stopper = threading.Event()

        def _worker():
            stopper.wait(0.1)

        thread = threading.Thread(target=_worker)
        thread.start()
        self.receiver.thread_vis = thread
        renderer = DummyRenderer()
        self.receiver.renderer = renderer
        self.receiver.stop()
        self.assertFalse(self.receiver.running_vis)
        self.assertTrue(renderer.closed)


if __name__ == "__main__":
    unittest.main()



import time
from collections import deque
from typing import Dict
from lunar_tools.logprint import dynamic_print

class FPSTracker:
    """A class to track and display FPS (Frames Per Second) with color-coded output and segment timing."""
    
    def __init__(self, buffer_size=30, update_interval=0.5):
        """
        Initialize the FPS tracker.
        
        Args:
            buffer_size (int): Number of frames to keep in rolling window for FPS calculation
            update_interval (float): How often to update the FPS display in seconds
        """
        self.fps_buffer = deque(maxlen=buffer_size)
        self.last_time = time.time()
        self.fps_update_interval = update_interval
        self.last_fps_update = time.time()
        self.current_fps = 0
        
        # Segment timing
        self.segments: Dict[str, float] = {}
        self.current_segment_start = None
        self.current_segment_name = None
        self._should_clear_segments = False
        
    def start_segment(self, name: str):
        """Start timing a new segment and end the previous one if it exists."""
        current_time = time.time()
        
        if self.current_segment_name is not None:
            # Record the previous segment's duration
            duration = current_time - self.current_segment_start
            self.segments[self.current_segment_name] = duration * 1000  # Convert to milliseconds
        
        self.current_segment_name = name
        self.current_segment_start = current_time
        
    def update(self):
        """
        Update FPS calculation. Call this method once per frame.
        Returns True if FPS display should be updated this frame.
        """
        current_time = time.time()
        
        # Handle the last segment if it exists
        if self.current_segment_name is not None:
            duration = current_time - self.current_segment_start
            self.segments[self.current_segment_name] = duration * 1000
            self.current_segment_name = None
            self.current_segment_start = None
        
        dt = current_time - self.last_time
        self.last_time = current_time
        self.fps_buffer.append(dt)
        
        should_update_display = current_time - self.last_fps_update >= self.fps_update_interval
        if should_update_display:
            if len(self.fps_buffer) > 0:
                self.current_fps = len(self.fps_buffer) / sum(self.fps_buffer)
            self.last_fps_update = current_time
            self._should_clear_segments = True
            
        return should_update_display
    
    def get_fps(self):
        """Get the current FPS value."""
        return self.current_fps
    
    def get_colored_fps_string(self):
        """Get a color-coded string representation of the current FPS."""
        # Clear the line first
        base_str = f"\033[2K\r\033[92mFPS: {self.current_fps:.1f}"  # Always green, with line clear
        
        # Add segment timings if any
        if self.segments:
            base_str += " | "
            segments_str = " | ".join(f"{name}: {ms:.1f}ms" for name, ms in self.segments.items())
            base_str += segments_str
            
        # Add a space at the end to ensure clean overwrite
        return base_str + " \033[0m"
    
    def print_fps(self):
        self.update()
        dynamic_print(self.get_colored_fps_string())
        if self._should_clear_segments:
            self.segments.clear()
            self._should_clear_segments = False 